#!/usr/bin/env python
# encoding: utf-8


"""
g_wave.py

Created by Chris Stevens 2022
Copyright (c) 2022 University of Canterbury. All rights reserved.
"""

# Import Python libraries
import numpy as np
from functools import partial

from coffee.tslices import tslices
from coffee.system import System
from coffee.diffop.fd import ghost_point_processor

# A simple bump function for the boundary condition
def bump(t):
	return np.sin(8.*t)**7.

# Class that describes a simple advection equation
class OneDAdvection(System):

	################################
	# Constructor
	################################

	def __init__(self, D, tau, CFL, char_speed):
		super(OneDAdvection, self).__init__()
		self.D          = D
		self.tau        = tau
		self.CFL        = CFL
		self.char_speed = char_speed
		self.name       = "1D Advection equation"
		self.numvar     = 1

	############################################################################
	# Boundary functions
	############################################################################
		
	# Left boundary condition
	def left(self, t):
		return bump(t)

	############################################################################
	# Evolution Routine
	############################################################################

	def evaluate(self, t, U, intStep = None):

		# Define useful variables
		Psi = U.data[0]
		dx  = U.domain.step_sizes[0]

		# Calculate spatial derivative
		DxPsi = self.D(Psi, dx)

		# Communicate spatial derivative through neighbouring MPI process(es)
		new_derivatives, _ = U.communicate(
			partial(ghost_point_processor),
			data=np.array([
				DxPsi
			])
		)
		DxPsi = new_derivatives[0]

		# Evaluate right hand side
		DtPsi = -self.char_speed*DxPsi

		# Impose boundary condition with SAT method
		pt_l       = self.D.penalty_boundary(dx, "left")
		pt_l_shape = pt_l.size
		C_left     = self.tau * pt_l * (-self.char_speed)
		l_Psi      = self.left(t)

		b_data = U.external_slices()
		for dim, direction, _ in b_data:
			if direction == 1:
				pass
			else:
				DtPsi[:pt_l_shape]  -= C_left * (Psi[0] - l_Psi)

		# Keep track of time
		print('t =', t)

		# Return RHS of evolution equation with SAT boundary condition imposed
		return tslices.TimeSlice([DtPsi], U.domain, time = t)

	############################################################################
	# Initial Data Routine
	############################################################################

	def initial_data(self, t0, grid):

		# Set initial data
		x   = grid.meshes[0]
		Psi = np.zeros_like(x)

		# Return Timeslice object
		return tslices.TimeSlice([Psi], grid, time = t0)

	############################################################################
	# Timestep Routine
	############################################################################

	def timestep(self, U):

		return self.CFL * U.domain.step_sizes[0] / self.char_speed


#!/usr/bin/env python
# encoding: utf-8
"""
The abstract base class (abc) for System objects. 

This class serves to specify
the interface that System classes are assumed to have. These assumptions are
made by virtually all classes that call the system. For example, ibvp.IBVP,
subclasses of solver.Solver, almost all actions.hdf_actions.SimOutType
subclasses.

While you don't need to actually implement all the methods below, it really is
recommended. You may get errors about the methods missing as a consequence.

system.System has abc.ABCMeta as a meta class. This serves as a warning to
really think about what you are doing if you don't implement all methods below.
"""

from abc import ABCMeta


#############################################################################
class System(object, metaclass=ABCMeta):
    """The System class that specifies the interface for all other system
    classes.

    You don't need to implement everything below, but you can expect
    errors if you don't. The main purpose of this class is to document
    the API that classes that describe a system of equations are expected
    to provide.
    """

    def timestep(self, tslice):
        """Returns a number giving the time step for the next iteration.

        Parameters
        ----------

        tslice : tslices.TimeSlice
                A time slice with the current data.

        Returns
        -------

        float
                A number giving the dt for which the next data will be
                calculated in the next iteration.
        """
        return NotImplementedError("You need to implement this function")

    def evaluate(self, t, tslice):
        """Returns the data needed by the solver.Solver subclass needed to
        calculate the values of the functions at the next time step.

        For example, in a IBVP problem using an RK sover this method will
        return the values of the derivatives of the functions being evolved.

        Yes I know it all sounds a bit vague. But really, this method is the
        heart and soul of coffee. You need to know what you're doing, how the
        solver is implemented and what data the system is meant to pass to the
        solver.

        Parameters
        ----------

        t : float
            The time at which the 'derivatives' need to be calculated. NOTE
            THAT THIS TIME MAY BE DIFFERENT FROM THE TIME STORED IN u. The time
            stored in u is the time for the values of the functions stored in
            u, not the time at which the derivatives of the functions stored in
            u are needed. In the case of RK methods t and u.time can be
            different. YOU HAVE BEEN WARNED.

        tslice : tslice.TimeSlice
            Contains the data from which the
            'evaluate' method is meant to calculate data from.

        Returns
        -------

        tslice.TimeSlice
            data needed for the solver.Solver subclass to calculate the
            values of the functions being evolved at the next time slice.
        """
        return NotImplementedError("You need to implement this function")

    def initialValues(self, t, grid):
        """Returns initial data for the simulation.

        Parameters
        ----------

        t : float
            The time at which the initial data is to be calcualted.
        grid : ABCGrid
            The grid over which the data is to be calculated.

        Returns
        -------
        tslice.TimeSlice
            The initial values for the functions to be evolved.
        """
        return NotImplementedError("You need to implement this function")

    def left(self, t):
        """Returns the boundary data on the 'left' boundary.

        Yes this also makes the assumption that left is well defined. See the
        documentation for free_data.FreeData. To be honest I'm not sure that
        this method needs to be defined here. After all: does any class other
        than the system class need to access this method? If not there there is
        no need for it.

        As no grid object is passed as an argument the boundary value must
        be independent of current function data. This is an obvious defect.
        Also: this doesn't seem to be a problem: boundary data can be manually
        handled in the evaluate routine. Still this issue should be changed
        when it actually becomes a problem.

        Parameters
        ----------

        t : float
            The time for which the left data needs to be calculated.

        Returns
        -------

            The boundary data.
        """
        return NotImplementedError("You need to implement this function")

    def right(self, t):
        """Returns the boundary data on the 'right' boundary.

        See the documentation for the function system.System.left

        Parameters
        ----------

        t : float
            The time for which the left data needs to be calculated.

        Returns
        -------

            The boundary data.

        """
        return NotImplementedError("You need to implement this function")



from builtins import range
import time

# from pylab import *
import pylab
from coffee.backend import backend as be

from coffee.actions import Prototype


class Plotter(Prototype):
    """An action that plots data using matplotlib."""

    def __init__(self, frequency=1, xlim=(-1, 1), ylim=(-1, 1), findex=None, delay=0.0):
        """Initialiser for Plotter.

        Parameters
        ==========
        frequency : int
            The number of iterations per execution of this action
        xlim : a two tuple of floats
            The limits of the x axes.
        ylim : a two tuple of floats
            The limits of the y axes.
        findex : int
            The number of functions to plot
        delay : float
            The delay, in seconds, between repeated executions of this function.
        """
        super(Plotter, self).__init__(frequency)
        if findex is not None:
            self.index = findex
            self.delay = delay
            self.colors = ("b", "g", "r", "c", "m", "y", "k", "coral")
            from numpy import asarray

            pylab.ion()
            fig = figure(1)
            ax = fig.add_subplot(111)
            self.lines = [ax.add_line(Line2D(xlim, ylim)) for k in findex]
            ax.set_xlim(xlim)
            ax.set_ylim(ylim)
            ax.grid(True)
            self.axes = ax

    def _doit(self, it, u):
        x = u.x
        index = be.asarray(self.index)
        f = u.fields[index]
        mx = be.max(f.flat)
        mn = be.min(f.flat)
        self.axes.set_title("Iteration: %d, Time: %f" % (it, u.time))
        # self.axes.set_xlim(x[0],x[-1])
        # self.axes.set_ylim(mn,mx,auto=True)

        l = len(self.index)
        pylab.ioff()
        for k in range(l):
            line = self.lines[k]
            line.set_xdata(x)
            line.set_ydata(f[k])
            line.set_color(self.colors[k])
        pylab.ion()
        pylab.draw()
        time.sleep(self.delay)



#!/usr/bin/env python
# encoding: utf-8
"""
The FreeData class describes the \`free data' of a system of equations.

This is a subclass of Abstract Base Class. It specifies an interface. All
classes that are used to calculate initial data or free data for a simulation
scheme are assumed to implement the methods below. In particular the object
actions.hdf_output.SimOutputType.Exact uses the exact method.

It is not, in general, needed to implement all of the following methods. What
you do implement should be determined by what additional objects interact with
your subclass of the system.System class.

This object is subclassed from abc inorder to make you think twice about what
you are doing. If your free data is sufficiently simple the code that
corresponds to free data could be included directly into system.System since
all calls to FreeData are mediated through system.System.
"""

import abc
from abc import ABCMeta

class FreeData(object, metaclass=ABCMeta):
    """This is a simple abstract base class for the implementation of exact
    solutions.

    All the 'virtual' functions must be defined, modulo the comments given in
    the module docstring.
    """

    @abc.abstractmethod
    def left_boundary(self, tslice):
        """Returns the boundary values given on the left boundary over a one dimensional grid.

        The currect FreeData object assumes that the simulation is being
        done over a one dimensional grid. Hence the call to 'left_boundary'.
        this method will eventually be altered to bring it into line with the
        grid method, \`boundary_slices' which specifies the slices of the domain
        corresponding to the boundary in a mannor which is independent of
        assumptions about the number of dimensions.

        Until then you'll just have to cope.

        Parameters
        ----------

        tslice : tslice.TimeSlice
            The timeslice of the data which left boundary values need to given.

        Returns
        -------

            The values of the functions being numerically simulated on
            left boundary. The type of these values is up to you,
            inconjunction with your implementation of the tslices.TimeSlice
            and system.System classes.
        """
        raise NotImplementedError("You must define the function left_boundary().")

    @abc.abstractmethod
    def right_boundary(self, tslice):
        """Returns the boundary values given on the right boundary over a one
        dimensional grid.

        Arguments and returned object are as for left_boundary.

        Parameters
        ----------

        tslice : tslice.TimeSlice
            The timeslice of the data which left boundary values need to given.

        Returns
        -------

            The values of the functions being numerically simulated on
              left boundary. The type of these values is up to you,
              inconjunction with your implementation of the tslices.TimeSlice
              and system.System classes.

        """
        raise NotImplementedError("You must define the function right_boundary().")

    @abc.abstractmethod
    def initial_data(self, t, grid):
        """Returns the initial data for a simulation based on the initial
        time and grid.

        Parameters
        ----------

        t : float
            The initial time at which the initial data is to be calculated.

        grid :
            The grid over which the initial data is to be calculated.

        Returns
        -------

        tslice.TimeSlice
            the initial data of the simulation.
        """
        raise NotImplementedError("You must define the function initial_data().")

    @abc.abstractmethod
    def exact(self, t, grid):
        """Returns the exact values of the functions being numerically found.

        The idea is that this data can be calculated and stored in the hdf file
        during the simulation as a result of a call to the
        actions.hdf_actions.SimOutputType.Exact object.

        Of course you don't have to use this method that way...

        Parameters
        ----------

        t : float
            The time at which to calculate the exact values.
        grid :
            The domain over which to calculate the exact values.

        Returns
        -------

            The exact values of the functions being solved for in the simulation.

        """
        raise NotImplementedError("You must define the function exact().")

    @abc.abstractmethod
    def dirichlet(self, u, intStep=None):
        """Return dirichlet boundary conditions.

        This is a poorly thought through addition to this class. The idea is
        that at certain steps during the evolution of data it might be
        necessary to call dirichlet boundary conditions. This method does that.
        Why is something different from left_boundary needed? I don't know.
        That's why it's poorly thought through. The bottom line is that this
        method might be removed from this class at some point.

        Originally this method was used to implement the intermediate boundary
        conditions for explicit RK routines during evolution. In order to do
        this knowledge of which intermediate step the boundary condition was
        being applied is necessary. This is the reason for the keyword
        argument.

        Parameters
        ----------

        u : tslice.TimeSlice
            Stores data for which dirichlet boundary conditions are needed.

        intStep : int
            Giving which intermediate step the RK routine is currently in.

        """
        raise NotImplementedError("You must define the function dirichlet().")



#!/usr/bin/env python
# encoding: utf-8
"""
This is the class that handles the iterative step and actions of the
simulation. 

The name ibvp comes from the original design of this class as the
iterative step in an initial boundary value problem solver. However, this class
is general enough to handle numerical techniques that have, at their highest
level, some iterative method.
"""

class IBVP:
    """Handles computation of the solution to initial boundary value problems.

    This class manages the interative process of simulations and interleaves
    interative steps with calls to a list of actions. Information about events
    during the simulation are writen to logging.getLogger("IBVP").
    """

    theActions = None
    theGrid = None
    theSolver = None
    iteration = 0
    maxIteration = None

    def __init__(self, sol, eqn, grid, action=[], maxIteration=10000, minTimestep=1e-8):
        """IBVP constructor.

        Parameters
        ----------

        sol : solver.Solver
              An ODE solver.

        eqn : system.System
              The system of equations whose solution will be computed.

        grid : grid.Grid
               The spatial grid over which the values of the solution to the system
               will be computed.

        action : list, actions.Prototype
                 Elements of the list must be instances of actions.Prototype. Each
                 action will be run after initialisation of the timeslice and after
                 each new timeslice is calculated.

        maxIteraction : int, Optional
                 Specifies the maximum number of timesteps.

        minTimestep : float, Optinoal
                 Specifies the minimum time step.
        """
        self.theSolver = sol
        self.theSystem = eqn
        self.maxIteration = maxIteration
        self.theGrid = grid
        self.theActions = action
        self.minTimestep = minTimestep

    def run(self, tstart, tstop=float("inf"), thits=None):
        """Go for it! Starts the simulation.

        Parameters
        ----------

        tstart : float
            A number giving the time at the initial interation.

        tstop : float
            A number giving the time to finish the simulation at.

        thits : list, float
            A list of times that the simulation is forced to hit exactly.

        """
        # Set up thits
        if thits is None:
            thits = []
        if tstop not in thits:
            thits += [tstop]

        # Order the list of times, and ensure that they are popped from smallest
        # to largest.
        thits = sorted(thits)
        thits.reverse()
        tstop = thits.pop()

        # Set start time.
        t = tstart

        # Get initial data and configure timeslices for multiple processors.
        u = self.theSystem.initial_data(t, self.theGrid)

        advance = self.theSolver.advance
        computation_valid = True
        while computation_valid and t < tstop:
            # Check against maxIteration
            if self.iteration > self.maxIteration:
                break

            dt = self.theSystem.timestep(u)

            # Check dt for size
            if dt < self.minTimestep:
                break

            # Check if dt needs to change in order to hit the next thits value.
            timeleft = tstop - t
            if timeleft < dt:
                dt = timeleft
                if not thits:
                    computation_valid = False
                else:
                    tstop = thits.pop()

            # Run the actions.
            self._run_actions(t, u)

            try:
                t, u = advance(t, u, dt)
            except OverflowError as e:
                # OverflowErrors arn't always appropirately handled. So we
                # do it ourselves here.
                print("Overflow error({0}): {1}".format(e.errno, e.strerror))
                computation_valid = False

            # Dirichlet boundary conditions
            try:
                u = self.theSystem.dirichlet_boundary(u)
            except AttributeError:
                pass

            # If we're using an mpi enable grid, this ensures that all
            # processes have gotten to the same point before continuing the
            # simulation.
            u.barrier()

            # On to the next iteration.
            self.iteration += 1

        # end (while)

        # Run the actions once more before exiting.
        self._run_actions(t, u)

        # This statement might be unnecessary. In principle it ensures that all
        # processes are about to complete the current simulation before exit
        # occurs.
        u.barrier()

        return u

    def _run_actions(self, t, u):
        """A utility method that ensures that actions are only run after all data is collected across MPI nodes and only if an action is due to run.

        Parameters
        ----------

        t : float
            The current time.

        u: tslices.TimeSlice
            The current timeslice.
        """

        # Ideally u.collect_data() should only be executed if there
        # actions that will run. Because of single process access to
        # actions this causes an issue.
        # Some thought is required to fix this problem.
        tslice = u.collect_data()
        if tslice is not None:
            actions_do_actions = [
                action.will_run(self.iteration, tslice) for action in self.theActions
            ]
            if any(actions_do_actions):
                for i, action in enumerate(self.theActions):
                    if actions_do_actions[i]:
                        action(self.iteration, tslice)


"""
Contains classes that abstract the MPI C API.

Classes in this module model the MPI node network and manage the relationship
of the complete domain to each 
"""
from builtins import range
from builtins import object
import abc
from coffee.backend import backend as be


###############################################################################
# Abstract Base Class for MPI Interfaces
###############################################################################
class MPIInterface(object, metaclass=abc.ABCMeta):
    def __init__(self, mpi_comm, *args, **kwds):
        """Initialisaer for MPIInterface objects.

        Parameters
        ==========
        mpi_comm: mpi4py.Comm
            A wrapper to an MPI_COMM object via mpi4py.
        """
        self.comm = mpi_comm
        super(MPIInterface, self).__init__(*args, **kwds)

    @abc.abstractmethod
    def subdomain(self):
        """Returns a tuple of slice objects that represents the subdomain of the full
        grid that this MPI node should perform computations over.

        Returns
        =======
        tuple of slices
        """

    @abc.abstractmethod
    def communicate(self, data):
        """Returns a list of tuples. The first element of each tuple is a slice
        representing the portion of data which corresponds to the second
        element of the tuple. The second element of the tuple is an array of
        data which are the communicated values that correspond to the
        positions described by the slice."""

    def barrier(self):
        if self.comm is None:
            return
        return self.comm.barrier()


###############################################################################
# Concrete implementations
###############################################################################
class EvenCart(MPIInterface):
    def __init__(self, domain, boundary_data, *args, **kwds):
        """Initialiser for EvenCart objects

        An EvenCart object is an implementation of MPIInterface that
        assumes that the global topology uses the MPI Cartesian methods.

        Parameters
        ==========
        domain: numpy.ndarray
            A numpy array representing the values of points in the global
            computational domain
        boundary_data: grid.ABCBoundary
        """
        super(EvenCart, self).__init__(*args, **kwds)
        self.domain = domain
        self.domain_mapping = self._make_domain_mappings(domain, boundary_data)

    def _make_domain_mappings(self, domain, boundary_data):
        """A utility that constructs the tuples of slices that represent
        the sub-grids for each MPI node based on that nodes rank.

        The calculation assumes that each MPI node should handle roughly the
        same load as all others and that the MPI topology is Cartesian.

        Parameters
        ==========
        domain: numpy.ndarray
            A numpy array representing the values of points in the global
            computational domain
        boundary_data: grid.ABCBoundary

        Returns
        =======
        list of tuples of slices:
            Each tuple of slices in index i gives the sub grid for MPI node
            with rank i.

        """
        if self.comm is None:
            return tuple([slice(None, None, None) for dim in domain])
        r_map = []
        for rank in range(self.comm.size):
            coords = self.comm.Get_coords(rank)
            r_map += [
                tuple(
                    [
                        self._array_slice(
                            domain[i],
                            coord,
                            self.comm.dims[i],
                            boundary_data.ghost_points(i, -1),
                            boundary_data.ghost_points(i, 1),
                        )
                        for i, coord in enumerate(coords)
                    ]
                )
            ]
        return r_map

    def _neighbour_slices(self, shape, b_data):
        """A utility function that collects the source and destination
        ranks for neighbouring MPI nodes along with the slices needed
        to describe what data should be transferred.

        Parameters
        ==========
        shape: tuple of ints
            The shape of the data that will be communicated to neighbouring MPI
            nodes

        b_data: grid.ABCBoundary

        Returns
        =======
        list of tuples:
            Each tuple has three members, first the source MPI node rank,
            second the destination MPI node rank and third the list of tuples
            of slices that describes the data to be communicated.
        """
        if self.comm is None:
            return []
        dims = self.comm.Get_dim()
        pos_neighbours = [
            b_data.source_and_dest(d, 1) + b_data.internal_slice(shape, d, 1)
            for d in range(dims)
        ]
        neg_neighbours = [
            b_data.source_and_dest(d, -1) + b_data.internal_slice(shape, d, -1)
            for d in range(dims)
        ]
        neighbours = pos_neighbours + neg_neighbours
        return neighbours

    def _array_slice(
        self, array_length, rank, num_ranks, ghost_points_start, ghost_points_end
    ):
        """The utility method that calculates the slices that describe the
        sub-grid managed by the MPI node with rank ``rank``.

        The method assumes that it is doing the calculation for a single
        dimension. To build the full list of tuples of slices, all
        dimensions will need to be iterated over.

        Parameters
        ==========
        array_length: int
            The number of grid points in this particular dimension of the
            total computational domain.
        rank: int
            The MPI rank that the slices are computed for.
        num_ranks: int
            The number of MPI nodes that cover this particular dimension.
        ghost_points_start: int
            The number of ghost points at the start of the dimension.
        ghost_points_end: int
            The number of ghost points at the end of the dimension.

        Returns
        =======
        slice:
            A slice which gives the subdomain for the particular dimension and
            the needed MPI node rank.
        """
        # divide domain into appropriate parts
        q, r = divmod(array_length, num_ranks)
        # use the rank to details which part is relevant for this process
        s = rank * q + min(rank, r)
        e = s + q
        # Adjust e to account for min(rank, r) term, which spreads the remainder
        # over the appropriate number of processes
        if rank < r:
            e = e + 1
        # add in ghost_points if we can
        # this currently works for gp = 1 for an SAT boudnary method.
        # The code will need to be checked for consistency if this changes.
        if s - ghost_points_start > -1:
            s = s - ghost_points_start
        if e + ghost_points_end < array_length:
            e = e + ghost_points_end
        return slice(s, e, None)

    @property
    def subdomain(self):
        """Returns the slices which give the subdomain for the MPI node
        that this method is called on.

        Returns
        =======
        list of tuples of slices
        """
        if self.comm is None:
            return self.domain_mapping
        return tuple(self.domain_mapping[self.comm.rank])

    def communicate(self, data, b_data):
        """Communicates the given data to appropriate neighbouring nodes.

        Parameters
        ==========
        data: numpy.ndarray
            An array representing the values of the computed function over
            the sub-grid managed by the MPI node this code is being computed on.
        b_data: grid.ABCBoundary

        Returns
        =======
        list of tuples (list of slices, numpy.ndarray)
            The returned data is a list of tuples. The first element of the tuple
            describes which grid points the communicated data relates to.
            The second element is the data received from whichever MPI node
            manages the sub-grid which looks after the relevant grid points.
        """
        # if self.comm.size == 1:
        # return []
        nslices = self._neighbour_slices(data.shape, b_data)
        r_data = []
        for source, dest, send_slice, recv_slice in nslices:
            if dest < 0:
                send_data = None
            else:
                # I really did not want to copy this view. For large data
                # sets this feels like an unnecessary delay.
                # However mpi4py did not seem to work with views and using
                # self.comm.sendrecv to send and recv arbitrary objects.
                # Plus self.comm.Sendrecv requires contiguous data
                # so only in rare cases is it possible to send.
                # Hence the be.copy statement is required.
                send_data = be.array(data[send_slice], copy=True, order="C")
            if source < 0:
                recv_data = None
            else:
                # when testing the above assertion I suggest changing this
                # line to be.ones_like, rather than be.empty_like
                recv_data = be.empty_like(data[recv_slice])
            self.comm.Sendrecv(send_data, dest=dest, recvbuf=recv_data, source=source)
            if source >= 0:
                r_data += [(recv_slice, recv_data)]
        return r_data

    def collect_data(self, data):
        """A method that collects all computed function values over all
        sub-grids managed by all MPI nodes and returns the amalgamated data
        if the rank of this process is 0.

        This gives a way to get a single copy of all computed data so far. It does,
        however, assume that all subsequence processing occurs on a single MPI
        node. This is a bottle neck.

        Parameters
        ==========
        data: numpy.ndarray
            The data to be communicated.

        Returns
        =======
        None or numpy.ndarray
            The collated data if MPI_Comm_Rank is 0, none otherwise.
        """

        # Note that this method does not take account of ghost_points in
        # the domains. This does not cause a problem. It just means
        # that more data than necessary is passed.

        # If there are no partitions of the domain we don't need to collect
        if self.comm is None:
            return data
        # Gather the data
        fields = self.comm.gather(data, root=0)
        # If this process is root then collate the data and return
        if self.comm.rank == 0:
            rdata_edims_shape = (data.shape[0],) + data.shape[len(self.domain) + 1 :]
            rdata_edims_slice = tuple(
                [slice(None, None, None) for i in rdata_edims_shape]
            )
            rdata = be.zeros(
                (rdata_edims_shape[0],) + self.domain + rdata_edims_shape[1:],
                dtype=data.dtype,
            )
            for rank, field in enumerate(fields):
                dslice = self.domain_mapping[rank]
                rdata[(rdata_edims_slice[0],) + dslice + rdata_edims_slice[1:]] = field
            return rdata
        else:
            return None



from builtins import object
from coffee.backend import backend as be


class Prototype(object):
    """The prototype of all actions.

    This class provides basic functionality that all actions require.

    To subclass this class:
    1) define the method _doit in the subclass
    2) call the Prototypes constructor to ensure that frequency, start and stop
    are properly initialised.
    """

    def __init__(
        self,
        frequency=1,
        start=-float("infinity"),
        stop=float("infinity"),
        thits=None,
        thits_toll=1e-14,
    ):
        """The initialiser for the Prototype action class.

        Parameters
        ==========
        frequency : int, Optional
            The action will only run after the given number of iterations.
        stop : float, Optional
            After this time the action will not run.
        start : float, Optional
            Before this time the action will not run.
        thits : list of floats, Optional
            Times at which the action is guaranteed to run
        thits_toll : float, Optional
            The tollerance for comparison against times in thits.
        """
        self.frequency = frequency
        self.stop = stop
        self.start = start
        if thits is not None:
            self.thits = be.asarray(thits)
        else:
            self.thits = None
        self.thits_toll = thits_toll

    def will_run(self, it, u):
        """Returns true if the action will run for the given iteration and
        time slice.

        Parameters
        ==========
        it : int
            The number of iterations in this simulation.
        u : tslice.TimeSlice
            The current timeslice.

        Returns
        =======
        boolean :
            True if the action will run.
        """
        test = (it % self.frequency) == 0 and self.start <= u.time <= self.stop
        if self.thits is not None:
            test = test and (be.absolute(u.time - self.thits) < self.thits_toll).any()
        return test

    def __call__(self, it, u):
        """If the action will run, runs the action.

        Parameters
        ==========
        it : int
            The number of iterations in this simulation.
        u : tslice.TimeSlice
            The current timeslice.
        """
        if self.will_run(it, u):
            self._doit(it, u)

    def _doit(self, it, u):
        pass


class BlowupCutoff(Prototype):
    """An action that stops the simulation if some component of the computed
    solution is above the given value.
    """

    def __init__(self, cutoff=10, **kwds):
        """The initialiser for the BlowupCutoff action.

        The parameters listed below are only the parameters specific to this
        action. All parameters as given for the initialiser for the
        actions. Prototype actions are also valid.

        Parameters
        ==========
        cutoff: float
            The value that is used to test for ending the simulation.
        """
        super(BlowupCutoff, self).__init__(**kwds)
        self.cutoff = abs(cutoff)

    def above_cutoff(self, u):
        """Returns true if the absolute value of any component of
        the data in the timeslice u is above the absolute value of
        the given cutoff.

        Parameters
        ==========
        u : tslice.TimeSlice
            The timeslice to be checked.

        Returns
        =======
        boolean :
            True if there is a component greater than the cutoff.
        """
        for component in u.fields:
            if be.any(be.abs(component) >= self.cutoff):
                return True
        return False

    def _doit(self, it, u):
        if self.above_cutoff(u):
            raise Exception("Function values are above the cutoff")


class Info(Prototype):
    """Logs the current iterations and data slice.

    Included more as an example than anything else.
    """

    def __init__(self, *args, **kwds):
        """The initialiser for the Info action.

        All parameters as given for the initialiser for the
        actions.Prototype action are valid.
        """
        super(Info, self).__init__(*args, **kwds)

    def _doit(self, it, u):
        pass



import logging
from unittest.mock import MagicMock

import pytest

from coffee.actions import actions
from coffee.backend import backend as be

backends = ["numpy"]
# When other backends are added, add them to this list
# e.g. backends = ["numpy", "cupy", "jax"]


@pytest.mark.parametrize("backend", backends)
class TestPrototype:
    def test_will_run(self, backend):
        be.set_backend(backend)
        proto = actions.Prototype(frequency=2, start=1, stop=3)
        u = MagicMock()

        u.time = 0.5
        assert not proto.will_run(0, u)
        u.time = 1.5
        assert proto.will_run(2, u)
        u.time = 2.5
        assert not proto.will_run(1, u)
        u.time = 3.5
        assert not proto.will_run(4, u)

    def test_call(self, backend):
        be.set_backend(backend)
        proto = actions.Prototype()
        proto.will_run = MagicMock(return_value=True)
        proto._doit = MagicMock()
        u = MagicMock()

        proto(1, u)
        proto.will_run.assert_called_once_with(1, u)
        proto._doit.assert_called_once_with(1, u)


@pytest.mark.parametrize("backend", backends)
class TestBlowupCutoff:
    def test_above_cutoff(self, backend):
        be.set_backend(backend)
        blowup = actions.BlowupCutoff(cutoff=10)
        u = MagicMock()

        u.fields = [be.array([5, 9]), be.array([9, -6])]
        assert not blowup.above_cutoff(u)

        u.fields = [be.array([5, 10]), be.array([10, 20])]
        assert blowup.above_cutoff(u)

    def test_doit(self, backend):
        be.set_backend(backend)
        blowup = actions.BlowupCutoff(cutoff=10)
        u = MagicMock()

        u.fields = [be.array([5, 9]), be.array([9, -6])]
        blowup._doit(1, u)  # No exception should be raised

        u.fields = [be.array([5, 10]), be.array([10, 20])]
        with pytest.raises(Exception, match="Function values are above the cutoff"):
            blowup._doit(1, u)


@pytest.mark.parametrize("backend", backends)
class TestInfo:
    def test_doit(self, backend, caplog):
        be.set_backend(backend)
        info = actions.Info()
        u = MagicMock()
        u.time = 2.0

        with caplog.at_level(logging.INFO, logger="Info"):
            info._doit(3, u)
        assert "Iteration: 3, Time: 2.000000" in caplog.text



#!/usr/bin/env python
# encoding: utf-8
"""
A module that contains dissipation operators for SBP operators.
"""

import math
import abc

################################################################################
# Base classes for SBP artificial dissipation operators
################################################################################
class Diss(object):
    """The parent class for SBP dissipation operators."""

    name = "Dissipation"

    def __init__(self, *args, **kwds):
        pass

    @abc.abstractmethod
    def __call__(self, u, dx, boundary_ID=None):
        pass

    def _consistancy_check(self, u, shape):
        r, c = shape
        if u.shape[0] + 1 <= 2 * c:
            raise ValueError("Domain too small for application of operator")

    def __str__(self):
        return "SBP Dissipation operator" % self.name


class DissDiag(Diss):
    """A class that represents dissipation operators that are diagonal."""

    name = "Dissipation Diagonal Norm"

    def __init__(self):
        self.bdyRegion = self.Ql.shape

    def __call__(self, u, dx, boundary_ID=None):
        super(DissDiag, self)._consistancy_check(u, self.bdyRegion)
        r, c = self.bdyRegion
        diss_u = be.zeros_like(u)
        diss_u = be.convolve(u, self.A, mode="same")
        if boundary_ID is None:
            diss_u[0:r] = be.dot(self.Ql, u[0:c])
            diss_u[-r:] = be.dot(self.Qr, u[-c:])
        elif boundary_ID == grid.LEFT:
            diss_u[0:r] = be.dot(self.Ql, u[0:c])
        elif boundary_ID == grid.RIGHT:
            diss_u[-r:] = be.dot(self.Qr, u[-c:])
        factor = math.pow(0.5, 2 * self.p)
        return factor * diss_u


class DissRestFull(Diss):
    """A class that represents dissipation operators for restricted full norms."""

    name = "Dissipation Restricted Full Norm"

    def __init__(self, p):
        """The Initialiser for this class.

        Parameters
        ----------
        p : int
            The power of the operator.
        """
        self.bdyRegion = self.Ql.shape
        self.p = p

    def __call__(self, u, dx, boundary_ID=None):
        """Applies the operator.

        We follow the notation of Diener Dorband Schnetter and Tiglio. We
        make the assumption that the differencing operator D_p is always of the
        form of some central finite difference stencil in the interior and some
        boundary matrix on the boundary.

        An important assumption is that D decomposes into the sum of two
        matrices A, Q. Where A represents the application of a centred
        finite difference operator on the interior and Q the application of the
        centred difference on the boundary.

        If the stencil for A is [-1,3, -3, 1] then it is assumed that
        D has the form:

            [[-1, 3, -3, 1, 0,...], [-1, 3, -3, 1, 0,...], [-1, 3, -3, 1, 0,...],
             [0, -1, 3, -3, 1, 0,...], [0, 0, -1, 3, -3, 1, 0, ...], ... ]

        and that the number of times the stencil is repeated on the boundary
        is the stencil length (4 in this case) minus 1 (to give 3 repeats).
        Hence Q is assumed to have the form:

            [[-1, 3, -3, 1, 0,...], [-1, 3, -3, 1, 0,...], [0,0,0,0,0,...],
             [0,0,0,0,0,0,...], ... ]

        and A is assumed to be D-Q. This ensures that if
        r,c = self.bdyRegion then r = self.A.shape[0]//2. This is an important
        assumption that is applied in the formula below to calculate
        Transpose[A].B.A.u where B is the non-constant diagonal matrix required
        to ensure numerical stability (see the paper).

        The explicit form of A and Q here ensures that Transpose[A].Q = 0
        and Transpose[Q].A = 0. This implies that
        Transpose[D].B.D.u = Transpose[Q].B.Q.u + Transpose[A].B.A.u
        This equation reduces the need for further calculations involving
        Transpose[A].Q and Transpose[Q].A

        In principle it is possible to remove these restrictions, i.e. allow
        r neq self.A.shape[0]//2 and to allow Transpose[A].Q neq 0
        and Transpose[Q].A neq 0, but the code below WILL need to be
        rewritten.

        Parameters
        ----------
        u: tslice.TimeSlice
            The data
        ds : float
            The spatial step size
        boundary_ID: int
            See the SBP operator module.

        Returns
        -------
        numpy.ndarray:
        """
        # This performs a consistance check
        super(DissRestFull, self)._consistancy_check(u, self.Ql.shape)

        # Get useful variables
        size = u.shape[0]
        r, c = self.Ql.shape

        # First do the internal convolution.
        # The array A is the central difference stencil that makes up the
        # "interior" portion of D_p. By assumption A is a matrix with the first
        # r rows full of zeros and the first non-zero row starting with the
        # stencil. That is if the stencil is [1, -2, 1] then the top left corner
        # of A is [[0,0,0,...], [0,0,0,...], [1, -2, 1, 0,...], ...
        # This implies that the convolvution operator used to implement matrix
        # multiplication must take account of the zero rows.
        # It turns out that this can be done by using the 'valid' mode in
        # the be.convolve method. This will reduce the size of the output by r
        # elements on the end and beginning of the output array. The values that
        # should be in these positions are all zero due to the r zero'd rows in
        # the matrix A.
        diss_u_int = be.zeros_like(u)
        diss_u_int[r:-r] = be.convolve(u, self.A, mode="valid")
        for i in range(size):
            diss_u_int[i] = self.B(i, dx, size) * diss_u_int[i]

        # To multiply my Transpose[A], requires some care. The zero'd rows now
        # become columns and the stencil is reversed. I make here the assumption
        # that the stencil is always symmetric, up to sign! For stencils with
        # odd length (p is even) the sign does not change, but for stencils with
        # even length (p is odd) the sign does change. For the stencil
        # [-1,3,-3,1] the matrix A is
        # [[0,0,0,0,0,...], [0,0,0,0,0,...], [-1, 3, -3, 1, 0,...],
        # [0, -1, 3, -3, 1, 0,...], ...]
        # Hence Transpose[A] is
        # [[0, 0, -1, 0,...], [0, 0, 3, -1, 0,...], [0, 0, -3, 3, -1, 0,...],
        # [0, 0, 1, -3, 3, -1, 0,...],... ]
        # This is incontrast to the A matrix generated by the stencil [1,-2,1]
        # where Transpose[A] is
        # [[0, 1, 0,...], [0, -2, 1, 0,...], [0, 1, -2, 1, 0,...], ...]
        # This is why the term math.pow(-1, p) is incorporated.
        #
        # On top of this be.convolve performs the convolution iterating from the
        # end of the stencil. This makes a difference when boundary affects of
        # convolution are desirable (as in this case). This is why there is
        # a ::-1 in the iteration for u in convolve. This did not need to be
        # accounted for in the calculation above because of the use of the
        #'valid' mode.
        #
        # Lastly the differences in vector length diss_u[r:-r] and diss_u is
        # inculded to account for the vector lengthening of the 'full' mode
        # of the convolve method which adds r number of terms to the vector
        # at both ends.
        diss_u_int = be.convolve(
            diss_u_int[r:-r], math.pow(-1, self.p) * self.A[::-1], mode="full"
        )

        # Second do the boundary convolution
        # The check at the beginning of the method ensures that the
        # two array's diss_u_b[0:r] and diss_u_b[-r:] do not share any points
        # of diss_u_b
        diss_u_b = be.zeros_like(u)
        if boundary_ID is None:
            diss_u_b[0:r] = be.dot(self.Ql, u[0:c])
            diss_u_b[-r:] = be.dot(self.Qr, u[-c:])
            for i in range(size):
                diss_u_b[i] = self.B(i, dx, size) * diss_u_b[i]
            diss_u_b[0:c] = be.dot(diss_u_b[0:r], self.Ql)
            diss_u_b[-c:] = be.dot(diss_u_b[-r:], self.Qr)
        elif boundary_ID == grid.LEFT:
            diss_u[0:r] = be.dot(self.Ql, u[0:c])
            for i in range(size):
                diss_u_b[i] = self.B(i, dx, size) * diss_u_b[i]
            diss_u_b[0:c] = be.dot(diss_u_b[0:r], self.Ql)
        elif boundary_ID == grid.RIGHT:
            diss_u[-r:] = be.dot(self.Qr, u[-c:])
            for i in range(size):
                diss_u_b[i] = self.B(i, dx, size) * diss_u_b[i]
            diss_u_b[-c:] = be.dot(diss_u_b[-r:], self.Qr)

        # Add the two parts together and multiply by the appropriate
        # numerical factor.
        diss_u = diss_u_int + diss_u_b
        factor = math.pow(0.5, 2 * self.p) * math.pow(dx, 2 * self.p - 2)
        diss_u = -factor * diss_u

        # Multiply by the inverse of the norm
        super(DissRestFull, self)._consistancy_check(u, self.norm_inv.shape)
        r, c = self.norm_inv.shape
        diss_u[0:r] = be.dot(self.norm_inv, diss_u[0:c])
        diss_u[-r:] = be.dot(self.norm_inv[::-1, ::-1], diss_u[-c:])

        # return the result
        return diss_u


################################################################################
# Dissipation for diagonal norm SBP operators.
################################################################################
class Diss21_DDST(DissDiag):
    """A dissipation operator.

    Taken from Diener, Dorband, Schnetter and Tiglio. Due to consistence
    conditions on SBP dissipation operators and the derivative operators this
    dissipation operator can only be used with the D21 operator from the same
    paper. Note that D21 is unique so the operator sbp.D21_CNG is the
    correct operator.
    """

    def __init__(self, *args, **kwds):
        self.A = be.array([1.0, -2.0, 1.0])
        self.name = "Diss21_DDST"
        self.p = 2

        self.Ql = be.zeros((2, 3))

        self.Ql[0, 0] = -2
        self.Ql[0, 1] = 2.0
        self.Ql[0, 2] = 0.0
        self.Ql[1, 0] = 1.0
        self.Ql[1, 1] = -2.0
        self.Ql[1, 2] = 1.0

        self.Qr = -self.Ql[::-1, ::-1]
        self.bdyRegion = (2, 3)


class Diss42_DDST(DissDiag):
    """A dissipation operator.

    Taken from Diener, Dorband, Schnetter and Tiglio. Due to consistency
    conditions on SBP dissipation operators and the derivative operators this
    disspation operator can only be used with the sbp.D42 operator which
    is from the same
    paper.
    """

    def __init__(self, *args, **kwds):
        self.A = be.array([-1.0, 4.0, -6.0, 4.0, -1.0])
        self.name = "Diss42_DDST"
        self.p = 4

        Q = be.zeros((4, 6))

        Q[0, 0] = -2.8235294117647058823529411764705882352941176470588
        Q[0, 1] = 5.6470588235294117647058823529411764705882352941176
        Q[0, 2] = -2.8235294117647058823529411764705882352941176470588
        Q[0, 3] = 0.0
        Q[0, 4] = 0.0
        Q[0, 5] = 0.0

        Q[1, 0] = 1.6271186440677966101694915254237288135593220338983
        Q[1, 1] = -4.0677966101694915254237288135593220338983050847458
        Q[1, 2] = 3.2542372881355932203389830508474576271186440677966
        Q[1, 3] = -0.81355932203389830508474576271186440677966101694915
        Q[1, 4] = 0.0
        Q[1, 5] = 0.0

        Q[2, 0] = -1.1162790697674418604651162790697674418604651162791
        Q[2, 1] = 4.4651162790697674418604651162790697674418604651163
        Q[2, 2] = -6.6976744186046511627906976744186046511627906976744
        Q[2, 3] = 4.4651162790697674418604651162790697674418604651163
        Q[2, 4] = -1.1162790697674418604651162790697674418604651162791
        Q[2, 5] = 0.0

        Q[3, 0] = 0.0
        Q[3, 1] = -0.97959183673469387755102040816326530612244897959184
        Q[3, 2] = 3.9183673469387755102040816326530612244897959183673
        Q[3, 3] = -5.8775510204081632653061224489795918367346938775510
        Q[3, 4] = 3.9183673469387755102040816326530612244897959183673
        Q[3, 5] = -0.97959183673469387755102040816326530612244897959184

        self.Ql = Q
        self.Qr = -self.Ql[::-1, ::-1]
        self.bdyRegion = (4, 6)
        super(Diss42_DDST, self).__init__()


################################################################################
# Dissipation for restricted full norm SBP operators.
################################################################################
class Diss43_DDST(DissRestFull):
    """A dissipation operator.

    Compatibility considerations require that this disspation operator only
    used with sbp.D43_Tiglioetal.
    """

    def __init__(self, bdy_percent):
        """Iniitialisation for Diss43_DDST.

        Parameters
        ----------
        bdy_percent : float
            A tuning parameter for the cut_off parameter used in the matrix
            B for numerical stability.
        """
        self.bdy_percent = bdy_percent
        self.p = 2

        self.A = be.array([1, -2, 1])
        self.Ql = be.array([[1, -2, 1]])
        self.Qr = self.Ql

        self.norm_inv = be.zeros((5, 5))
        self.norm_inv[0, 0] = 4.186595370392226897362216859769846226369
        self.norm_inv[0, 1] = 0
        self.norm_inv[0, 2] = 0
        self.norm_inv[0, 3] = 0
        self.norm_inv[0, 4] = 0
        self.norm_inv[1, 0] = 0
        self.norm_inv[1, 1] = 0.6725191921225620731888714836983116420871
        self.norm_inv[1, 2] = 0.3613418181134949259370502966736306984367
        self.norm_inv[1, 3] = -0.2021316117293899791481674539631879662707
        self.norm_inv[1, 4] = 0.03455320708729270824077678274955265350304
        self.norm_inv[2, 0] = 0
        self.norm_inv[2, 1] = 0.3613418181134949259370502966736306984367
        self.norm_inv[2, 2] = 0.7206133711630147057720442098623847362950
        self.norm_inv[2, 3] = 0.1376472340546569368321616389764958792591
        self.norm_inv[2, 4] = -0.04136405531324488624637892257286207044784
        self.norm_inv[3, 0] = 0
        self.norm_inv[3, 1] = -0.2021316117293899791481674539631879662707
        self.norm_inv[3, 2] = 0.1376472340546569368321616389764958792591
        self.norm_inv[3, 3] = 0.9578653607931026822074133441449909689509
        self.norm_inv[3, 4] = 0.02069353627247161734563597102894256809696
        self.norm_inv[4, 0] = 0
        self.norm_inv[4, 1] = 0.03455320708729270824077678274955265350304
        self.norm_inv[4, 2] = -0.04136405531324488624637892257286207044784
        self.norm_inv[4, 3] = 0.02069353627247161734563597102894256809696
        self.norm_inv[4, 4] = 0.9908272703370861473007798925906968380654

        super(Diss43_DDST, self).__init__(self.p)

    def B(self, i, dx, size):
        """Returns an element of the matrix `B` which ensures numerical stability.

        Parameters
        ----------
        i: int
            The row and column, of the diagonal matrix, to return
        dx: float
            The step size
        size: int
            Length of the data
        """
        cut_off = math.floor(size * self.bdy_percent)
        if i < cut_off:
            return dx + (i / cut_off) * (1 - dx)
        elif i > size - 1 - cut_off:
            return 1 + (cut_off + 1 - size + i) * (dx - 1) / cut_off
        else:
            return 1


#!/usr/bin/env python
# encoding: utf-8
"""
This module contains the abstract base class for solvers as well as several
default implementations.

This default implementations all assume that tslice.TimeSlice objects can be
operated on by numerically (e.g. addition, multiplication, subtraction).
"""


from builtins import object
import abc

from coffee.tslices import tslices


################################################################################
# Solver Abstract Base Class
################################################################################
class ABCSolver(object):
    """The ABSolver abstract base class.

    This class provides the interface that is assumed of all other solvers.
    These methods are called in the ibvp.IBVP class.
    """

    def __init__(self, system, *args, **kwds):
        """Returns an instance of ABCSolver.

        This method should be called by any subclasses. It defines the system
        atribute and sets up a logging object.

        Parameters
        ----------

        system : system
            The system being used.
        """
        self.system = system
        super(ABCSolver, self).__init__(**kwds)

    @abc.abstractproperty
    def name(self):
        """This should return a name associated to the subclass.

        The name will need to be defined in the subclasses constructor. The
        name is used to identify the instantiated class in the logger.

        Returns
        -------

        string
            A string name for the class. This defaults to ABCSolver.

        """
        return "ABCSolver"

    @abc.abstractmethod
    def advance(self, t, u, dt):
        """Returns a tslice.TimeSlice containing data at time t + dt.

        Parameters
        ----------

        t : float
            The current time.

        u : tslice.TimeSlice
            The current tslice.TimeSlice. Note that the time in the timeslice
            is the time for the data in the timeslice, but not necessarily
            equal to t.

        dt : float
            The step in time required.

        Returns
        -------

        tslice.TimeSlice
            Contains the data at time t+dt.

        """
        raise NotImplementedError("This method needs to be implemented")

    def __repr__(self):
        return "<%s: system = %s>" % (self.name, self.system)


################################################################################
# First order methods
################################################################################
class Euler(ABCSolver):
    """An implementation of the first order Euler method."""

    name = "Euler"

    def advance(self, t, u, dt):
        """An Euler method implementation of ABCSolver.

        See the ABCSolver.advance method for documentation.
        """
        du = self.system.evaluate(t, u)
        r_time = t + dt
        r_slice = u + dt * du
        r_slice.time = r_time
        return (r_time, r_slice)


class ImplicitEuler(ABCSolver):
    """An implementation of the first order implicit Euler method."""

    name = "ImplicitEuler"

    def _NR(self, u_start, u_next, t, dt, domain):
        """A Newton-Raphson auxilliary routine for
        the implementation of the implicit Euler scheme.

        Parameters
        ----------
        u_start: numpy.ndarray
            Initial values of the function.

        u_next: numpy.ndarray
            The first guess at the next values of the function.

        t: float
            Current time.

        dt: float
            Current time step.

        domain: grid
            The domain that the function is being evaluated over.

        """

        # Starting guess for Newton-Raphson algorithm. Currently
        # chooses current tslice.
        # To be perhaps moved so user specifies this in system file or
        # perhaps in setup file when choosing implicit Euler.
        prev_u_next = u_next

        while True:
            # Evolution equation of u
            g = self.system.evaluate(
                t + dt, tslices.TimeSlice(prev_u_next, domain, time=t)
            ).data

            # Derivative of evolution equation of u w.r.t. evolved variable
            dg, TOL = self.system.implicit_method_jacobian(
                t + dt, tslices.TimeSlice(prev_u_next, domain, time=t)
            )

            dg = dg.data

            # NR is applied to f
            f = prev_u_next - u_start - dt * g
            df = 1.0 - dt * dg

            # Compute next value of u_next
            next_u_next = prev_u_next - (f / df)

            # If we get to within a certain accuracy of the root stop
            if all(abs(next_u_next - prev_u_next)[0] < TOL):
                break

            # Otherwise continue
            prev_u_next = next_u_next

        return next_u_next

    def advance(self, t, u, dt):
        """An implicit Euler method implementation of ABCSolver.

        See the ABCSolver.advance method for documentation.
        """

        # Current values of the evolved variables
        u_start = u.data

        # Initial guess for the Newton-Raphson method is the current value(s)
        # of the evolved variables
        u_next = u_start

        # prev_u_next = u_next
        prev_u_next = u_next

        # Returns u_next from applying Newton-Raphson
        u_ret = self._NR(u_start, u_next, t, dt, u.domain)

        r_time = t + dt
        r_slice = tslices.TimeSlice(u_ret, u.domain, r_time)
        return (r_time, r_slice)


################################################################################
# Fourth order methods
################################################################################
class RungeKutta4(ABCSolver):
    """A RungeKutta4 implementation of ABCSolver."""

    name = "RK4"

    def __init__(self, *args, **kwds):
        super(RungeKutta4, self).__init__(*args, **kwds)

    def advance(self, t0, u0, dt):
        """See the ABCSolver.advance method for documentation.

        Very simple minded implementation of the standard 4th order Runge-Kutta
        method to solve an ODE of the form fdot = rhs(t,f)
        """
        eqn = self.system
        u = u0
        t = t0
        k = eqn.evaluate(t, u)
        u1 = u0 + (dt / 6.0) * k

        u = u0 + dt / 2.0 * k
        t = t0 + dt / 2.0
        k = eqn.evaluate(t, u)
        u1 += dt / 3.0 * k

        u = u0 + dt / 2.0 * k
        t = t0 + dt / 2.0
        k = eqn.evaluate(t, u)
        u1 += dt / 3.0 * k

        u = u0 + dt * k
        t = t0 + dt
        k = eqn.evaluate(t, u)
        u1 += dt / 6.0 * k
        u1.time = t

        return (t, u1)


class RungeKutta4Dirichlet(ABCSolver):
    """An implementation of the Runge Kutta 4 routine that calls
    system.dirichlet_boundary."""

    def __init__(self, **kwds):
        super(RungeKutta4Dirichlet, self).__init__(**kwds)
        if not hasattr(self.system, "dirichlet_boundary"):
            raise Exception(
                "%s does not implement dirichlet_boundary method" % self.system
            )

    def advance(self, t0, u0, dt):
        """See the ABCSolver.advance method for documentation.

        Very simple minded implementation of the standard 4th order Runge-Kutta
        method to solve an ODE of the form
        fdot = rhs(t,f) that allows for the implementation of
        Dirichlet conditions during evaluation.

        Ensure that the corresponding system file has a method called,
        "dirichlet_boundary".
        """
        eqn = self.theEqn
        u = u0
        t = t0
        k = eqn.evaluate(t, u, intStep=1)
        u1 = u0 + (dt / 6.0) * k
        u1 = eqn.dirichlet_boundary(u1, intStep=1)

        u = u0 + dt / 2.0 * k
        t = t0 + dt / 2.0
        k = eqn.evaluate(t, u, intStep=2)
        u1 += dt / 3.0 * k
        u1 = eqn.dirichlet_boundary(u1, intStep=2)

        u = u0 + dt / 2.0 * k
        t = t0 + dt / 2.0
        k = eqn.evaluate(t, u, intStep=3)
        u1 += dt / 3.0 * k
        u1 = eqn.dirichlet_boundary(u1, intStep=3)

        u = u0 + dt * k
        t = t0 + dt
        k = eqn.evaluate(t, u, intStep=None)
        u1 += dt / 6.0 * k
        u1.time = t
        u1 = eqn.dirichlet_boundary(u1, intStep=None)

        return (t, u1)



"""
An action is a class which performs some additinonal processing to a time_slice
before each iteration.

Please see the help for the prototype action for implementation details.
"""

from coffee.actions.actions import Prototype, BlowupCutoff, Info
from coffee.actions.hdf_output import SimOutput



"""A module containing actions that interface with GnuPlot."""


from builtins import str
from builtins import range
import time
import os
from coffee.backend import backend as be

# Gnuplot currently only works with Python 2!
try:
    import Gnuplot
except ImportError:
    import sys

    print(sys.exc_info())
    print("Gnuplot currently only works with Python 2!")
    sys.exit()

from coffee.actions import Prototype


class Plotter1D(Prototype):
    """This class provides convient access to GnuPlot's functions for one
    dimensional data.

    It is not
    suitable for animations since the GnuPlot object is persistent between
    calls to this class.

    The simulation itself does not need to be one dimensional. The data to be
    plotted is 'pre-processed' by a function that is passed by the user. In
    this way the user can, for example, specify if the modulus, or argument
    of complex data is to be plotted.
    """

    def __init__(self, system, *args, **kwds):
        """The constructor for Plotter1D objects.

        Plotter1D subclasses actions.Prototype.

        All positional arguments, excluding the
        first (``system``), are passed to the GnuPlot object. These
        arguments should be strings respecting the documentation of GnuPlot.
        They are passed to the Gnuplot object g as g(pos_argument)

        Additional parameters valid for actions.Prototype are valid.

        Parameters
        ==========
        system : system
            The system object.
        args : list of strings
            Passed to gnuplot as arguments.
        delay : float, Optional
            The delay in seconds between plotting events.
        title : string, Optional
            The title of the plot. The default is the time of the plot.
        data_function : function(it, u, system) -> (be.ndarray, be.ndarray), Optional
              A function that takes the interation number,
              a timeslice and the system. It should return a tuple consisting
              of the domain of the data and a two dimensional array. The first
              dimension of the array is considered to be the components. The
              second dimension is considered to give the domain dependence.
        """

        if "start" in kwds:
            start = kwds.pop("start")
        else:
            start = -float("Infinity")
        if "frequency" in kwds:
            frequency = kwds.pop("frequency")
        else:
            frequency = 1
        if "delay" in kwds:
            self.delay = kwds.pop("delay")
        else:
            self.delay = 0.0
        self.title = r"Time %f"
        if "title" in kwds:
            self.title = kwds.pop("title")
        self.system = system
        try:
            if kwds["data_function"] is not None:
                self.datafunc = kwds.pop("data_function")
            else:
                self.datafunc = lambda y, x, z: (x.domain.axes[0], x.data)
        except:
            self.datafunc = lambda y, x, z: (x.domain.axes[0], x.data)
        super(Plotter1D, self).__init__(frequency=frequency, start=start, **kwds)
        self.Device = Gnuplot.Gnuplot()
        g = self.Device
        for arg in args:
            g(arg)

    def _doit(self, it, u):
        g = self.Device
        x, f = self.datafunc(it, u, self.system)
        f = be.atleast_2d(f)
        graphs = []
        g("set title '" + self.title % u.time + "'")
        g.plot(*graphs)
        time.sleep(self.delay)

    def __del__(self):
        del self.Device


class Plotter2D(Prototype):
    """This class provides convient access to GnuPlot's functions for two
    dimensional data.

    It is not
    suitable for animations since the GnuPlot object is persistent between
    calls to this class.

    The simulation itself does not need to be two dimensional. The data to be
    plotted is 'pre-processed' by an function that is passed by the user.
    """

    def __init__(self, *args, **kwds):
        """The constructor for Plotter2D objects.

        Plotter2D subclasses actions.Prototype. Any argument valid
        in the initialiser of actions.Prototype is respected here.

        All positional arguments
        are passed to the GnuPlot object. These
        arguments should be strings respecting the documentation of GnuPlot.
        They are passed to the Gnuplot object g as g(pos_argument)

        The data_function
        should return a tuple consisting
        of the domain of the data and a three dimensional array. The first
        dimension of the array is considered to be a number of components
        each of which is to be plotted. The
        second and third dimensions is considered to give the domain dependence.

        Parameters
        ==========
        system : system
            The system object.
        args : list of strings
            Passed to gnuplot as arguments.
        delay : float, Optional
            The delay in seconds between plotting events.
        components : list of strings, Optional
            The strings are used to label the components being plotted.
        title : string, Optional
            The title of the plot. The default is the time of the plot.
        data_function : function(it, u, system) -> (be.ndarray, be.ndarray), Optional
              A function that takes the interation number,
              a timeslice and the system. It should return a tuple consisting
              of the domain of the data and a two dimensional array. The first
              dimension of the array is considered to be the components. The
              second dimension is considered to give the domain dependence.
        """
        if "delay" in kwds:
            self.delay = kwds.pop("delay")
        else:
            self.delay = 0.0

        if "title" in kwds:
            self.title = kwds.pop("title")
        else:
            self.title = "Iteration %d, Time %f"

        if "components" in kwds:
            self.components = kwds.pop("components")
        else:
            self.components = None

        self.system = kwds.pop("system")
        try:
            if kwds["data_function"] is not None:
                self.datafunc = kwds.pop("data_function")
            else:
                self.datafunc = lambda y, x, z: (x.domain.axes, x.data)
        except:
            self.datafunc = lambda y, x, z: (x.domain.axes, x.data)

        self.device = Gnuplot.Gnuplot()
        for arg in args:
            self.device(arg)
        super(Plotter2D, self).__init__(**kwds)

    def _doit(self, it, u):
        x, f = self.datafunc(it, u, self.system)
        f = be.atleast_2d(f)
        graphs = []
        self.device('set title "%s" enhanced' % self.title % (it, u.time))
        for i in range(f.shape[0]):
            if self.components is None:
                graphs += [
                    Gnuplot.GridData(
                        f[i, :, :],
                        xvals=x[0],
                        yvals=x[1],
                        filename="deleteme.gp",
                        title="Component %i" % i,
                        binary=0,
                    )
                ]
            else:
                graphs += [
                    Gnuplot.GridData(
                        f[i, :, :],
                        xvals=x[0],
                        yvals=x[1],
                        filename="deleteme.gp",
                        title=self.components[i],
                        binary=0,
                    )
                ]
        self.device.splot(*graphs)
        time.sleep(self.delay)

    def __del__(self):
        try:
            os.remove("deleteme.gp")
        except:
            print(
                "The file deleteme.gp was unable to be deleted. Please do so manually."
            )
        del self.device



#!/usr/bin/env python
# encoding: utf-8


"""
g_wave.py

Created by Chris Stevens 2022
Copyright (c) 2022 University of Canterbury. All rights reserved.
"""

import numpy as np
from functools import partial

from coffee.tslices import tslices
from coffee.system import System
from coffee.diffop.fd import ghost_point_processor

rt2 = np.sqrt(2.)
r2 = 1./np.sqrt(2.)

def bump(t):
	return np.sin(8.*t)**7.

class OneDAdvection(System):

	################################
	# Constructor
	################################

	def __init__(self, D, tau, CFL, char_speed):
		super(OneDAdvection, self).__init__()
		self.D          = D
		self.tau        = tau
		self.CFL        = CFL
		self.char_speed = char_speed
		self.name       = "1D Advection equation"
		
	def left(self, t):
		return bump(t)

	def evaluate(self, t, U, intStep = None):

		# Define useful variables
		Psi = U.data[0]

		x  = U.domain.axes[0]
		dx = U.domain.step_sizes[0]

		# Calculate spatial derivative
		DxPsi = self.D(Psi, dx)

		# Evaluate right hand side
		DtPsi = -self.char_speed*DxPsi

		# Impose boundary conditions
		new_derivatives, _ = U.communicate(
			partial(ghost_point_processor),
			data=np.array([
				DtPsi
			])
		)
		DtPsi = new_derivatives[0]

		pt_l       = self.D.penalty_boundary(dx, "left")
		pt_l_shape = pt_l.size

		C_left  = self.tau * pt_l * (-self.char_speed)

		l_Psi = self.left(t)

		b_data = U.external_slices()
		for dim, direction, d_slice in b_data:
			if direction == 1:
				pass
			else:
				DtPsi[:pt_l_shape]  -= C_left * (Psi[0] - l_Psi)

		print('t =', t)

		return tslices.TimeSlice([DtPsi], U.domain, time = t)

	def initial_data(self, t0, grid):

		x   = grid.meshes[0]
		Psi = np.zeros_like(x)

		tslice = tslices.TimeSlice([Psi], grid, time = t0)
		return tslice

	def timestep(self, U):

		return self.CFL * U.domain.step_sizes[0]


"""The io module provides a wrapper around h5py so as to translate between
system, solver, and timeslice objects in a COFFEE friendly way.

Datagroups in an hdf file are assumed to contain all data associated to a 
simulation. Each dataset in a data group 

The module hdf_output is closely related to this module. The dictionary
dgTypes maps data group names to SimOutputType class names. It is best
to read the hdf_output module documentation as well as this modules
documentation if the dgTypes dictionary needs to be altered.
"""

from builtins import zip
from builtins import map
from builtins import str
from builtins import range
from builtins import object
import functools
import h5py
from coffee.backend import backend as be

# Gnuplot currently only works with Python 2!
try:
    import Gnuplot
except ImportError:
    pass

NUMERICAL_TOLERANCE = 1e-14
# Important configuration is included after these two classes


# A wrapper class that helps ease iteration over datasets with
# str(int) indices going from 0,1,... upwards.
class DataGroup(object):
    """The DataGroup class wraps a h5py group so that the setter, getter
    and iter methods do sensible array-like things.

    Datagroups are assumed to represent an entire simulation. Datasets
    are assumed to represent each timeslice. Each dataset in a data group
    is the value of a keyword "i" where i is the iteration number of the
    timeslice that the dataset contains.

    This class abstracts these details so that a given a data group ``g`` we
    can write ``g[2]'' to get the timeslice of the second iteration
    and ``g[3]`` to get the timeslice of the following iteration.
    """

    global group

    @property
    def attrs(self):
        """Wraps the attrs dictionary of a hdf datagroup."""
        return self.group.attrs

    @property
    def name(self):
        """Wraps the hdf datagroup name."""
        return self.group.name

    def attrs_list(self, kwd):
        """Represents the attributes of each data set of the data group
        as a list.

        Returns
        =======
        list :
            Each entry in the list is a value of some keyword in the
            attrs dictionary of a sub data set.
        """
        list = []
        for data_set in self:
            list += data_set.attrs[kwd]
        return list

    def index_of_attr(
        self, attr, value, start_index=0, value_comparor=lambda x: x == value
    ):
        """Returns the index of the dataset in self whose attribute attr
        has the given value.

        The function value_comparor allows for fudging a
        little. It is a function that takes the attrs of a data set and
        returns true or false. Once true is found the index of that data
        set is returned.

        Parameters
        ==========
        attr :
            A keyword to be applied to the attrs attribute of each dataset.
        value :
            The value that is searched for.
        start_index : int, Optional
            The index at which to start the search.
        value_comparor : function(value)
            A function which returns true if the given value matched the
            desired value.

        Returns
        =======
        int :
            The index of the dataset whose attrs attributes matched.
        """
        index = -1
        for i in range(start_index, len(self)):
            if value_comparor(self[i].attrs[attr]):
                index = self[i].attrs["index"]
                break
        return index

    def __init__(self, grp, returnValue=False):
        """The group to behave like an array. It is assumed that
        the group has/will have a number of datasets with the labels
        '0','1', etc...

        Parameters
        ==========
        grp : h5py.Group
            The data group that this object will wrap.
        returnValue : bool
            If true __getitem__ will return the value of the dataset rather
            than the dataset itself.
        """
        self.group = grp
        self.rV = returnValue

    def __iter__(self):
        """Iterates in increasing numerical order 0,1,2,3,...
        across the datasets of the group. Returns the dataset
        at position 0,1,2,3...

        A generator that allows for iterative access to datasets.

        Yields
        ======
        h5py.dataset :
        """
        i = 0
        while True:
            try:
                yield self[i]
                i += 1
            except:
                return

    def __len__(self):
        """Returns the number of datasets.

        Returns
        =======
        int :
            The number of datasets in this group.
        """
        return len(self.group)

    def __setitem__(self, i, value):
        """Allows specification of the value of a dataset at a given index.

        Parameters
        ==========
        i : int
            The index at which value should be placed
        value :
            The value of the data set at index i. Must be an object that
            h5py can place into a dataset.
        """
        value = be.array(value)
        dataset = self.group.require_dataset(str(i), value.shape, value.dtype)
        dataset[:] = value
        dataset.attrs["index"] = i

    def __getitem__(self, i):
        """Returns either the dataset or the value at index i.

        Parameters
        ==========
        i : int
            The index of the dataset to be returned.

        Returns
        =======
        dataset or value :
            The value will be returned only if the returnValue variable is
            true
        """
        if self.rV:
            return self.group[str(i)][()]
        return self.group[str(i)]

    def __repr__(self):
        return r"<H5pyArray datagroup %s (%d)>" % (self.name, len(self))


class DomainDataGroup(DataGroup):
    """A DataGroup wrapper that handles the specific case of function data
    over the computational grid.
    """

    def __setitem__(self, i, value):
        value = be.array(value)
        dataset = self.group.require_dataset(str(i), value.shape, value.dtype)
        dataset[:] = value
        dataset.attrs["index"] = i

    def __getitem__(self, i):
        dataset = self.group[str(i)]
        axes_shape = dataset.attrs["axes_shape"]
        axes = []
        start = 0
        for i in range(len(axes_shape)):
            axes += [dataset[()][start : start + axes_shape[i]]]
            start = start + axes_shape[i]
        return axes

    def __repr__(self):
        return r"<H5pyArray domaindatagroup %s (%d)>" % (self.name, len(self))


# Do not change the keys!
#
# Each dgTypes (DataGroup Types) describes the group name for a data group
# structure in the hdf file. Each data group stores hdf datasets named by
# an index. The index across different dgtypes gives data for the corresponding
# iteration.
#
# The keys represent the names of the different dgtypes and are hard coded
# throughout the code. Therefore if you wish to rename a dgtype, just change
# the item, not the key.
#
# These represent the types of data that a simulation knows about.

dgTypes = {
    "raw": "Raw_Data",
    "constraints": "Constraints",
    "exact": "Exact_Data",
    "errorNum": "Error_Numeric",
    "errorExa": "Error_Exact",
    "IJ": "Weyl_Constants_IJ",
    "domain": "Domain",
    "time": "Time",
    "dt": "Time_Step",
    "scrif": "Scri+",
    "constraint": "Constraint",
    "mu": "mu",
    "mup": "mup",
}

"""The dgTypes dictionary maps data types that are produced during simulation
to the keys used in the hdf file to store that data. The dgTypesInv 
dictionary provide the reverse mapping.

If you need a new data type to be written out you can dynamically modify this
dictionary. Changes will also need to be reflected in the hdf output action."""

dgTypesInv = dict(list(zip(list(dgTypes.values()), list(dgTypes.keys()))))


dgTypes_DataGroups = {"domain": (None, DomainDataGroup)}

"""The dgTypes_DataGroups dictionary maps keys in the dgTypes dictionary to
a 2-tuple. 

The first entry in the two tuple is a module name that is imported
using  __import__, the second entry is the name of a class in that module that
provides DataGroup functionality suitable for the given dgType.

If the first entry is none then it is assumed that the appropriate data group
is in this module. The class name rather than the class string should be in
the second entry of the tuple.

If a dgType key is not in the dictionary the DataGroup class is used.
"""

# SystemDataTypes stores a list of all the subgroups in the system groups.

systemD = "System"

"""The systemD variable store the hdf key used to create datagroups that
store data for system objects."""

sysDTypes = {
    "system": systemD,
    "solver": "Solver",
    "grid": "Grid",
    "cmp": "cmp",
    "numvar": "NumVariables",
}

"""The sysDTypes dictionary lists the data types in a system object (keys)
against the key used in the hdf file to store that data (values).

Feel free to dynamically alter the dictionary. Logic is based on correctness
of the keys not the values. New data types for system objects should be added
in this dictionary as key value pairs.
"""


# An interface to ease interaction with the simulationHDF class when
# only a specific simulation is wanted. I expect this class to be used the
# most.
@functools.total_ordering
class Sim(object):
    """
    Represents the data associated to a simulation as stored in an HDF file.

    Designed to be accessed via a SimulationHDF object.
    """

    def __init__(self, simName, simHDF):
        """The initialiser for Sim objects.

        Parameters
        ==========
        simName : string
            The name of the simulation this object will represent.
        simHDF : SimulationHDF
            The SimulationHDF object that wraps the hdf file.
        """
        self.simHDF = simHDF
        self.name = simName
        existing_items = list(self.simHDF[systemD + "/" + self.name].keys())
        for key, item in list(sysDTypes.items()):
            if item in existing_items:
                setattr(self, key, self.simHDF[systemD + "/" + self.name][item][()])
        self.cmp = float(self.cmp)
        self.numvar = int(self.numvar)
        existing_items = list(self.simHDF.file.keys())
        for key, item in list(dgTypes.items()):
            if item in existing_items:
                if self.name in list(self.simHDF[item].keys()):
                    if key in dgTypes_DataGroups:
                        if dgTypes_DataGroups[key][0] is not None:
                            mod = __import__(
                                dgTypes_DataGroups[key][0],
                                fromlist=[dgTypes_DataGroups[key][1]],
                            )
                            dataGroup = getattr(mod, dgTypes_DataGroups[key][1])
                            setattr(
                                self,
                                key,
                                dataGroup(
                                    self.simHDF[item + "/" + self.name],
                                    returnValue=True,
                                ),
                            )
                        else:
                            setattr(
                                self,
                                key,
                                dgTypes_DataGroups[key][1](
                                    self.simHDF[item + "/" + self.name],
                                    returnValue=True,
                                ),
                            )
                    else:
                        setattr(
                            self,
                            key,
                            DataGroup(
                                self.simHDF[item + "/" + self.name], returnValue=True
                            ),
                        )
        setattr(
            self,
            "indices",
            sorted(
                map(
                    int,
                    list(
                        self.simHDF[
                            list(self.simHDF.file.keys())[0] + "/" + self.name
                        ].keys()
                    ),
                )
            ),
        )

    def tslice(self, i):
        """Wraps the SimulationHDF method of the same name.

        Assumes that the simulation used is the one represented by this object.

        Parameters
        ===========
        i : int
            The iteration number of the timeslice to return.
        sim: string
            The name of the simulation to access
        dgType : string, Optional
            The name of the type of data to retrieve. See the dgTypes dictionary
            for a list of possible values.

        Returns
        =======
        tslice.TimeSlice:
        """
        return self.simHDF.tslice(i, self.name, dgType=dgType["raw"])

    def indexOfTime(self, t):
        """Wraps the SimulationHDF method of the same name.

        Assumes that the simulation used is the one represented by this object.

        Parameters
        ==========
        t : float
            The time whose index is desired.
        sim : string
            The name of the simulation to search.

        Returns
        =======
        int:
            The iteration index that matches the given time up to the
            assumed NUMERICAL_TOLERANCE.
        """
        return self.simHDF.indexOfTime(t, self.name)

    def __eq__(self, other):
        """Rich comparison based on the cmp parameter.

        self.cmp is given by the comparison parameter specified during simulation.

        Returns
        =======
        bool:
        """
        return self.cmp == other.cmp

    def __lt__(self, other):
        """Rich comparison based on the cmp parameter.

        self.cmp is given by the comparison parameter specified during simulation.

        Returns
        =======
        bool:
        """
        return self.cmp < other.cmp

    def __str__(self):
        return self.name

    def write(self, dgType, it, data, name=None, derivedAttrs=None):
        """Wraps the SimulationHDF file of the same name.

        Assumes that the simulation name is given by this object.

        Parameters
        ==========
        dgType : string
            A key from the module level dgTypes dictionary.
        it : int
            The iteration number of the data to be written.
        data :
            The data to write. It must be able to be stored in an h5py.dataset.
        name : string, Optional
            A parameter used to create a sub-datagroup. See comments above.
        derivedAttrs : dictionary, Optional
            A dictionary of additional attributes to add to created datasets
        """
        self.simHDF.write(dgType, self.name, it, data, name, derivedAttrs)

    def getDgType(self, dgType):
        """Wraps the SimulationHDF method of the same name.

        Parameters
        ==========
        dgType: string
            A key from the dgTypes dictionary.

        Returns
        =======
        DataGroup:
        """
        return self.simHDF.getDgType(dgType, self.name)

    def getDgTypeAttr(self, dgType, attr, i):
        """Wraps the SimulationHDF class' method of the same name.

        Parameters
        ==========
        dgType : string
            A key of the dgTypes dictionary. The dgType whose attributes are desired.
        attr : string
            The attribute to return.
        i : int
            The index of the iteration whose attributes are desired.

        Returns
        =======
        DataGroup :
        """
        return self.simHDF.getDgTypeAttr(dgType, attr, i, self.name)

    def animate(
        self,
        dgType="raw",
        gnuCommands=None,
        gnuInitialisationCommands={"debug": 0, "persist": 1},
        tstart=-float("Infinity"),
        tstop=float("Infinity"),
        animationLength=2,
        framesPerSec=60,
    ):
        """A utility method that provide nice defaults for the GNUplot method.

        Parameters
        ==========
        dgType : string
            A key of the dbTypes dictionary. The type of data to be displayed.
        gnuCommands : list of strings, Optional
            A list of string commands for GnuPlot.
        gnuInitialisationCommands : list of strings, Optional
            A list of string commands to be parsed to GnuPlot on initialisation.
        tstart : float, Optional
            The time at which to start plotting data.
        tstop : float, Optional
            The time at which to stop plotting data.
        animationLength : float
            The length of time of the animation.
        framesPerSec : int
            The number of frames per second of display.
        """
        self.GNUplot(
            self.getDgType(dgType),
            gnuCommands=gnuCommands,
            gnuInitialisationCommands=gnuInitialisationCommands,
            tstart=tstart,
            tstop=tstop,
            animationLength=animationLength,
            framesPerSec=framesPerSec,
        )

    def plot(
        self,
        time,
        dgType="raw",
        gnuCommands=None,
        gnuInitialisationCommands={"debug": 0, "persist": 1},
    ):
        """A utility that plots the data at a particular time.

        This is a wrapper to the animate method that gives the same
        start and stop times.

        Parameters
        ==========
        time : float
            The time of the data that should be plotted.
        dgType : string, Optinal
            A key of the dgTypes dictionary.
        gnuCommands : list of strings
            Commands to be parsed to GnuPlot when the data is plotted.
        gnuInitialisationCommands : list of strings
            Commands to be parsed to GnuPlot on initialisation.
        """
        self.animate(
            dgType=dgType,
            gnuCommands=gnuCommands,
            gnuInitialisationCommands=gnuInitialisationCommands,
            tstart=time,
            tstop=time,
        )

    def GNUplot(
        self,
        group,
        gnuCommands=None,
        gnuInitialisationCommands=None,
        tstart=-float("Infinity"),
        tstop=float("Infinity"),
        animationLength=2,
        framesPerSec=60,
    ):
        """A utility function which plots a given group.

        There is no return, but gnuplot is called hence, depending on what
        parameters are parsed, plots will be created.

        Parameters
        ==========
        group : DataGroup
            A DataGroup containing the data to be plotted.
        gnuCommands : list of strings, Optional
            A list of string commands for GnuPlot.
        gnuInitialisationCommands : list of strings, Optional
            A list of string commands to be parsed to GnuPlot on initialisation.
        tstart : float, Optional
            The time at which to start plotting data.
        tstop : float, Optional
            The time at which to stop plotting data.
        animationLength : float
            The length of time of the animation.
        framesPerSec : int
            The number of frames per second of display.
        """
        if gnuCommands is None:
            gnuCommands = []
        if gnuInitialisationCommands is None:
            gnuInitialisationCommands = []

        # Get x values
        domains = self.getDgType("domain")

        # Get all times
        times = self.getDgType("time")
        times.rV = True
        numOfFrames = len(times)
        frameSkip = int(numOfFrames / (animationLength * framesPerSec))

        # Get data for scri
        scrif = self.getDgType("scrif")
        scrif.rV = True

        # Initialize gnuplot
        gnu = Gnuplot.Gnuplot(**gnuInitialisationCommands)
        gnu.reset()
        for command in gnuCommands:
            gnu(command)

        # Iterate across group
        # Get starting and stoping index
        nextFrame_index = self.indexOfTime(tstart)
        stop_index = self.indexOfTime(tstop)

        # While there is a next frame...
        while nextFrame_index < numOfFrames:
            # plot data
            i = nextFrame_index
            y = group[i]
            gnu.title("Simulation %s at time %f" % (self.name, times[i]))
            plotItems = []
            for j, row in enumerate(be.atleast_2d(y[()])):
                plotItems += [
                    Gnuplot.Data(domains[i], row, title="Component " + str(j))
                ]
            plotItems += [Gnuplot.Data(domains[i], scrif[i], title="Scri+")]
            gnu.plot(*plotItems)
            # if there are not enough frames left set the frameSkip to 0
            if nextFrame_index + frameSkip >= numOfFrames:
                frameSkip = 0
            nextFrame_index += 1 + frameSkip
            # if the next frame is larger than the stop_index then stop
            # plotting
            if nextFrame_index > stop_index:
                break
        gnu.close()

    class dsReturnValue(object):
        def __init__(self, dataset):
            self.ds = dataset

        def __getitem__(self, key):
            return self.ds[key][()]


# Allows for interaction with the hdf file without specific reference
# to a particular simulation. I expect that this class will only be used
# for easy access to the sim objects.
class SimulationHDF(object):
    """
    Represents the data associated to all simulations as stored in an HDF file.
    """

    def __init__(self, fileName, **kwds):
        """The initialiser for SimulationHDF.

        Parameters
        ==========
        filename: string
            The name of the hdf file to be wrapped.
        """
        self.file = h5py.File(fileName, "r+")

    def __enter__(self):
        return self

    def __exit__(self, type, value, traceback):
        self.file.close()

    def sim(self, name):
        """Returns a Sim class wrapping the data for the appropriate simulation."""
        return Sim(name, self)

    def name(self):
        """The name of the file that this class wraps.

        Returns
        =======
        string :
            The filename.
        """
        return self.file.name

    def getSims(self):
        """Returns a list of Sim objects representing the simulations contained
        in the hdf file.

        Returns
        =======
        list of Sim classses:
            The Sim classes wrapping the simulation contained in this hdf file.
        """
        simArray = [self.sim(name) for name in list(self.file[systemD].keys())]
        return sorted(simArray)

    def __getitem__(self, key):
        """Provides direct access to the underlying data group with the given key.

        Returns
        =======
        h5py.Group:
        """

        return self.file[key]

    def getSimData(self, sim):
        """Returns a dictionary giving access to the hdf objects corresponding
        to the given simulation name.

        Parameters
        ==========
        sim : string
            The name of the simulation whose data to retrieve.

        Returns
        =======
        dictionary:
        """
        sg = self.file[dgTypes["system"] + sim]
        rl = {"name": sim}
        for key, item in systemData:
            rl[key] = sg[item][()]
        return rl

    def tslice(self, i, sim, dgType="raw"):
        """The timeslice of the given simulation at the given iteration.

        Parameters
        ===========
        i : int
            The iteration number of the timeslice to return.
        sim: string
            The name of the simulation to access
        dgType : string, Optional
            The name of the type of data to retrieve. See the dgTypes dictionary
            for a list of possible values.

        Returns
        =======
        tslice.TimeSlice:
        """
        rdata = self.file[dgTypes[dgtype] + sim][str(i)]
        time = self.file[dgTypes["times"] + sim][str(i)]
        domain = self.file[dgTypes["domains"] + sim][str(i)]
        return tslice(rdata, domain, time)

    def getDgType(self, dgType, sim):
        """Returns a DataGroup object wrapping the given simulations data
        for the given dgType.

        Parameters
        ==========
        dgType: string
            A key from the dgTypes dictionary.
        sim: string
            The name of the simulation whose data to access.

        Returns
        =======
        DataGroup:
        """
        return DataGroup(self.file[dgTypes[dgType] + "/" + sim])

    def getDgTypeAttr(dgType, attr, i, sim):
        """Returns a DataGroup object wrapping the given simulations data set
        attributes for the given dgType.

        Parameters
        ==========
        dgType: string
            A key from the dgTypes dictionary.
        sim: string
            The name of the simulation whose data to access.

        Returns
        =======
        DataGroup:
        """
        return DataGroup(self.file[dgTypes[dgType] + "/" + sim]).attr[attr]

    def indexOfTime(self, t, sim):
        """A utility method that supports the translation of a simulation time
        to the iteration index.

        The tolerance for the required float comparison is given in
        the module variable NUMERICAL_TOLERANCE.

        Parameters
        ==========
        t : float
            The time whose index is desired.
        sim : string
            The name of the simulation to search.

        Returns
        =======
        int:
            The iteration index that matches the given time up to the
            assumed NUMERICAL_TOLERANCE.
        """
        times_dg = DataGroup(self.file[dgTypes["time"] + "/" + sim])
        indices = sorted([int(index) for index in list(times_dg.group.keys())])
        # If only one index
        if len(indices) == 1:
            if abs(times_dg[indices[0]][(0)] - t) <= NUMERICAL_TOLERANCE:
                return indices[0]
            else:
                return -1

        # Check initial step
        time_dg = times_dg[indices[0]]
        if indices[1] - indices[0] == 1:
            dt = times_dg[indices[1]][()] - time_dg[()]
            if t <= time_dg[()] < t + (dt / 2.):
                return indices[0]
        else:
            if abs(time_dg[(0)] - t) <= NUMERICAL_TOLERANCE:
                return indices[0]

        # Check all other steps except final
        for i in range(1, len(indices) - 1):
            time_dg = times_dg[indices[i]]
            if indices[i + 1] - indices[i] == 1:
                dt = times_dg[indices[i + 1]][()] - time_dg[()]
                if t - (dt / 2.) <= time_dg[()] < t + (dt / 2.):
                    return indices[i]
            else:
                if abs(time_dg[(0)] - t) <= NUMERICAL_TOLERANCE:
                    return indices[i]

        i = len(indices) - 1
        time_dg = times_dg[indices[i]]
        if indices[i] - indices[i - 1] == 1:
            dt = time_dg[()] - times_dg[indices[i - 1]]
            if t - (dt / 2.) <= time_dg[()] <= t:
                return indices[i]
        else:
            if abs(time_dg[(0)] - t) <= NUMERICAL_TOLERANCE:
                return indices[i]
        return -1

    def write(
        self, dgType, sim, it, data, name=None, derivedAttrs=None, overwrite=True
    ):
        """This method allows for writing to SimulationHDF objects.

        Note that if
        dgType is an error type data group then name must be given. We recommend
        that its value be taken as the data group from which the error data was
        generated.
        To ensure that name is not / is needed refer to how the data is
        extracted.

        Parameters
        ==========
        dgType : string
            A key from the module level dgTypes dictionary.
        sim : string
            The name of the simulation to write to.
        it : int
            The iteration number of the data to be written.
        data :
            The data to write. It must be able to be stored in an h5py.dataset.
        name : string, Optional
            A parameter used to create a sub-datagroup. See comments above.
        derivedAttrs : dictionary, Optional
            A dictionary of additional attributes to add to created datasets
        overwrite: bool
            If true data will be overwritten, otherwise an error will be
            raised.
        """
        # Create empy derivedAttrs if no argument is passed
        if derivedAttrs is None:
            self.derivedAttrs = {}
        else:
            self.derivedAttrs = derivedAttrs

        # If dgType is an error type then name must be set.
        # if dgType == sd.dgTypes["errorNum"] or\
        #    dgType == sd.dgTypes["errorExa"]:
        #    if name is None:
        #        raise Exception("""If SimulationHDF.write() is based an error dgType the name keyword must be set. Suggested usage is that name = the dgType of the data from which the error data was calculated.""")

        # get name if not none
        if name is not None:
            dg_name = dgType + "/" + sim + "/" + name
        else:
            dg_name = dgType + "/" + sim

        # get dg
        if overwrite:
            if name is not None:
                dg = DataGroup(
                    self.file.require_group(dgType)
                    .require_group(sim)
                    .require_group(name)
                )
            else:
                dg = DataGroup(self.file.require_group(dgType).require_group(sim))
        else:
            if name is not None:
                dg = DataGroup(
                    self.file.create_group(dgType).create_group(sim).create_group(name)
                )
            else:
                dg = DataGroup(self.file.create_group(dgType).create_group(sim))

        # add data and derived attrs
        dg[it] = data
        for key, value in list(self.derivedAttrs.items()):
            dg[it].attrs[key] = value


def array_value_index_mapping(correct, comparison, compare_on_axes=1):
    """A utility function which is useful when comparing data.

    The function takes two arrays and returns a list of pairs
    of indices (index1, index2) so that correct[index1] = comparison[index2]
    (up to NUMERICAL_TOLERANCE)
    this is very useful when performing error calculations.

    The idea is that data over two arbitrary domains can be easily compared.

    Parameters
    ==========
    correct : numpy.ndaray
    comparison : numpy.ndarray
    compare_on_axes : int
        If 0 then the first axes is included in the indices

    Returns
    =======
    list of 2-tuples of tuples of ints:
        Each pair of tuples of ints represents the data values which are
        defined over the same point.
    """
    index_mapping = []
    cor_dims = len(correct.shape)
    com_dims = len(comparison.shape)
    if compare_on_axes == 0:
        cor_ind = [0 for i in range(cor_dims)]
        com_ind = [0 for i in range(com_dims)]
    else:
        cor_ind = [0 for i in range(cor_dims - 1)]
        com_ind = [0 for i in range(com_dims - 1)]
    return _array_value_index_mapping_recursive(
        correct, cor_ind, comparison, com_ind, index_mapping, compare_on_axes, 0
    )


def _array_value_index_mapping_recursive(
    correct, cor_ind, comparison, com_ind, index_mapping, compare_on_axes, depth
):
    while (
        cor_ind[depth] < correct.shape[depth]
        and com_ind[depth] < comparison.shape[depth]
    ):
        if compare_on_axes == 0:
            com = comparison[tuple(com_ind)]
            cor = correct[tuple(cor_ind)]
        else:
            com = comparison[tuple(com_ind)][depth]
            cor = correct[tuple(cor_ind)][depth]
        if com + NUMERICAL_TOLERANCE < cor:
            com_ind[depth] = com_ind[depth] + 1
        elif com > NUMERICAL_TOLERANCE + cor:
            cor_ind[depth] = cor_ind[depth] + 1
        elif abs(com - cor) < NUMERICAL_TOLERANCE:
            if depth == len(correct.shape) - 1 - compare_on_axes:
                index_mapping += [(tuple(cor_ind), tuple(com_ind))]
                com_ind[depth] = com_ind[depth] + 1
                cor_ind[depth] = cor_ind[depth] + 1
            else:
                index_mapping = _array_value_index_mapping_recursive(
                    correct,
                    cor_ind,
                    comparison,
                    com_ind,
                    index_mapping,
                    compare_on_axes,
                    depth + 1,
                )
                com_ind[depth] = com_ind[depth] + 1
                cor_ind[depth] = cor_ind[depth] + 1
                com_ind[depth + 1] = 0
                cor_ind[depth + 1] = 0
        else:
            raise Exception("Unable to compare %s and %s" % (com, cor))
    return index_mapping



#!/usr/bin/env python
# encoding: utf-8

"""
Methods that wrap common computations
"""


class Backend:
    def __init__(self):
        self.backend_name = None
        self._backend = None

    def set_backend(self, backend_name="numpy"):
        self.backend_name = backend_name
        self._backend = self._get_backend(backend_name)

    def _get_backend(self, backend_name):
        """
        Returns an instance of the specified backend class.

        This function creates and returns an instance of the backend class specified by the
        `backend_name` argument. Currently supported backends are "numpy" and others can be
        added as needed.

        Args:
            backend_name (str, optional): The name of the backend for which an instance is to be created.
                Defaults to "numpy".

        Raises:
            ValueError: If the specified `backend_name` is not supported.

        Returns:
            object: An instance of the specified backend class.
        """
        backends = {
            "numpy": NumpyBackend,
            # Add other backends here as needed
        }

        if backend_name not in backends:
            raise ValueError(f"Unsupported backend: {backend_name}")

        return backends[backend_name]()

    def abs(self, *args, **kwargs):
        return self._backend.abs(*args, **kwargs)

    def absolute(self, *args, **kwargs):
        return self._backend.absolute(*args, **kwargs)

    def any(self, *args, **kwargs):
        return self._backend.any(*args, **kwargs)

    def apply_along_axis(self, *args, **kwargs):
        return self._backend.apply_along_axis(*args, **kwargs)

    def apply_over_axes(self, *args, **kwargs):
        return self._backend.apply_over_axes(*args, **kwargs)

    def arange(self, *args, **kwargs):
        return self._backend.arange(*args, **kwargs)

    def array(self, *args, **kwargs):
        return self._backend.array(*args, **kwargs)

    def array_equal(self, *args, **kwargs):
        return self._backend.array_equal(*args, **kwargs)

    def asarray(self, *args, **kwargs):
        return self._backend.asarray(*args, **kwargs)

    def atleast_1d(self, *args, **kwargs):
        return self._backend.atleast_1d(*args, **kwargs)

    def atleast_2d(self, *args, **kwargs):
        return self._backend.atleast_2d(*args, **kwargs)

    def convolve(self, *args, **kwargs):
        return self._backend.convolve(*args, **kwargs)

    def diag(self, *args, **kwargs):
        return self._backend.diag(*args, **kwargs)

    def dot(self, *args, **kwargs):
        return self._backend.dot(*args, **kwargs)

    def dtype(self, *args, **kwargs):
        return self._backend.dtype(*args, **kwargs)

    def empty(self, *args, **kwargs):
        return self._backend.empty(*args, **kwargs)

    def empty_like(self, *args, **kwargs):
        return self._backend.empty_like(*args, **kwargs)

    def exp(self, *args, **kwargs):
        return self._backend.exp(*args, **kwargs)

    def fromiter(self, *args, **kwargs):
        return self._backend.fromiter(*args, **kwargs)

    def lib_stride_tricks_as_strided(self, *args, **kwargs):
        return self._backend.lib_stride_tricks_as_strided(*args, **kwargs)

    def linspace(self, *args, **kwargs):
        return self._backend.linspace(*args, **kwargs)

    def log(self, *args, **kwargs):
        return self._backend.log(*args, **kwargs)

    def log2(self, *args, **kwargs):
        return self._backend.log2(*args, **kwargs)

    def mat(self, *args, **kwargs):
        return self._backend.mat(*args, **kwargs)

    def max(self, *args, **kwargs):
        return self._backend.max(*args, **kwargs)

    def min(self, *args, **kwargs):
        return self._backend.min(*args, **kwargs)

    def ndarray(self, *args, **kwargs):
        return self._backend.ndarray(*args, **kwargs)

    def nonzero(self, *args, **kwargs):
        return self._backend.nonzero(*args, **kwargs)

    def ones(self, *args, **kwargs):
        return self._backend.ones(*args, **kwargs)

    def ones_like(self, *args, **kwargs):
        return self._backend.ones_like(*args, **kwargs)

    def power(self, *args, **kwargs):
        return self._backend.power(*args, **kwargs)

    def savetxt(self, *args, **kwargs):
        return self._backend.savetxt(*args, **kwargs)

    def set_printoptions(self, *args, **kwargs):
        return self._backend.set_printoptions(*args, **kwargs)

    def squeeze(self, *args, **kwargs):
        return self._backend.squeeze(*args, **kwargs)

    def sum(self, *args, **kwargs):
        return self._backend.sum(*args, **kwargs)

    def tan(self, *args, **kwargs):
        return self._backend.tan(*args, **kwargs)

    def union1d(self, *args, **kwargs):
        return self._backend.union1d(*args, **kwargs)

    def vectorize(self, *args, **kwargs):
        return self._backend.vectorize(*args, **kwargs)

    def where(self, *args, **kwargs):
        return self._backend.where(*args, **kwargs)

    def zeros(self, *args, **kwargs):
        return self._backend.zeros(*args, **kwargs)

    def zeros_like(self, *args, **kwargs):
        return self._backend.zeros_like(*args, **kwargs)


class BackendBase:
    def __init__(self):
        pass


# ################################################################################
# # NumPy implementation
# ################################################################################


class NumpyBackend(BackendBase):
    def __init__(self):
        # print("Initialising Numpy Backend")
        try:
            import numpy as np
        except ImportError:
            raise ImportError(
                "NumPy library is not installed. Please install NumPy to use this function."
            )

        self.np = np

    def abs(self, *args, **kwargs):
        return self.np.abs(*args, **kwargs)

    def absolute(self, *args, **kwargs):
        return self.np.absolute(*args, **kwargs)

    def any(self, *args, **kwargs):
        return self.np.any(*args, **kwargs)

    def apply_along_axis(self, *args, **kwargs):
        return self.np.apply_along_axis(*args, **kwargs)

    def apply_over_axes(self, *args, **kwargs):
        return self.np.apply_over_axes(*args, **kwargs)

    def arange(self, *args, **kwargs):
        return self.np.arange(*args, **kwargs)

    def array(self, *args, **kwargs):
        return self.np.array(*args, **kwargs)

    def array_equal(self, *args, **kwargs):
        return self.np.array_equal(*args, **kwargs)

    def asarray(self, *args, **kwargs):
        return self.np.asarray(*args, **kwargs)

    def atleast_1d(self, *args, **kwargs):
        return self.np.atleast_1d(*args, **kwargs)

    def atleast_2d(self, *args, **kwargs):
        return self.np.atleast_2d(*args, **kwargs)

    def convolve(self, *args, **kwargs):
        return self.np.convolve(*args, **kwargs)

    def diag(self, *args, **kwargs):
        return self.np.diag(*args, **kwargs)

    def dot(self, *args, **kwargs):
        return self.np.dot(*args, **kwargs)

    def dtype(self, *args, **kwargs):
        return self.np.dtype(*args, **kwargs)

    def empty(self, *args, **kwargs):
        return self.np.empty(*args, **kwargs)

    def empty_like(self, *args, **kwargs):
        return self.np.empty_like(*args, **kwargs)

    def exp(self, *args, **kwargs):
        return self.np.exp(*args, **kwargs)

    def fromiter(self, *args, **kwargs):
        return self.np.fromiter(*args, **kwargs)

    def lib_stride_tricks_as_strided(self, *args, **kwargs):
        return self.np.lib_stride_tricks_as_strided(*args, **kwargs)

    def linspace(self, *args, **kwargs):
        return self.np.linspace(*args, **kwargs)

    def log(self, *args, **kwargs):
        return self.np.log(*args, **kwargs)

    def log2(self, *args, **kwargs):
        return self.np.log2(*args, **kwargs)

    def mat(self, *args, **kwargs):
        return self.np.mat(*args, **kwargs)

    def max(self, *args, **kwargs):
        return self.np.max(*args, **kwargs)

    def min(self, *args, **kwargs):
        return self.np.min(*args, **kwargs)

    def ndarray(self, *args, **kwargs):
        return self.np.ndarray(*args, **kwargs)

    def nonzero(self, *args, **kwargs):
        return self.np.nonzero(*args, **kwargs)

    def ones(self, *args, **kwargs):
        return self.np.ones(*args, **kwargs)

    def ones_like(self, *args, **kwargs):
        return self.np.ones_like(*args, **kwargs)

    def power(self, *args, **kwargs):
        return self.np.power(*args, **kwargs)

    def savetxt(self, *args, **kwargs):
        return self.np.savetxt(*args, **kwargs)

    def set_printoptions(self, *args, **kwargs):
        return self.np.set_printoptions(*args, **kwargs)

    def squeeze(self, *args, **kwargs):
        return self.np.squeeze(*args, **kwargs)

    def sum(self, *args, **kwargs):
        return self.np.sum(*args, **kwargs)

    def tan(self, *args, **kwargs):
        return self.np.tan(*args, **kwargs)

    def union1d(self, *args, **kwargs):
        return self.np.union1d(*args, **kwargs)

    def vectorize(self, *args, **kwargs):
        return self.np.vectorize(*args, **kwargs)

    def where(self, *args, **kwargs):
        return self.np.where(*args, **kwargs)

    def zeros(self, *args, **kwargs):
        return self.np.zeros(*args, **kwargs)

    def zeros_like(self, *args, **kwargs):
        return self.np.zeros_like(*args, **kwargs)

    def lib_stride_tricks_as_strided(self, x, shape=None, strides=None, writeable=True):
        return self.np.lib.stride_tricks.as_strided(
            x, shape=shape, strides=strides, writeable=writeable
        )
